import java.io.IOException;
import java.io.DataInputStream;
import java.io.InputStream;

/**
 * RRGGL Virtal Machine
 *
 * The class processes compiled script code. You need connect an interface contains codes of instructions that was generated by the compiler. 
 *
 * @author Igor Maznitsa (igor.maznitsa@igormaznitsa.com)
 * @version 1.00
 *
 * (C) 2004-2006 Raydac Research Group Ltd. (http://www.coldcore.ru)
 *
 * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in
 * the documentation and/or other materials provided with the distribution.
 * Neither the name of the RRGGL, IGORMAZNITSA.COM, IGOR MAZNITSA,RAYDAC RESEARCH GROUP, COLD CORE, RRG nor the names of its contributors may be used to endorse or promote
 * products derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
 * OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
public class RRGGLbinVM implements <PLACE HERE THE GENERATED HEADER NAME>
{
    /**
     * The size of the Arithmetical Stack
     */
    public static final int SIZE_SP = 256;

    /**
     * The size of the Return Stack
     */
    public static final int SIZE_RS = 512;

    /**
     * The index of Command Pointer (PC) register
     */
    public static final int REGISTER_PC = 0;

    /**
     * The index of Depth SP register
     */
    public static final int REGISTER_SP = 1;

    /**
     * The index of Depth RS register
     */
    public static final int REGISTER_RS = 2;

    //#if DEBUG
    /**
     * The debuf info data length
     */
    private static final int DEBUG_INFO_LENGTH = 5;
    //#endif

    /**
     * The array contains the data stack of the VM
     */
    protected int[] ai_sp;
    /**
     * The array contains the return stack of the VM
     */
    protected int[] ai_rs;
    /**
     * The array contains array area of the VM
     */
    protected int[] ai_arrays;

    /**
     * The array contains array dimension data
     */
    protected short[] ash_arrayDimensions;

    /**
     * The array contains arrays offset data in the common data block
     */
    protected int[] ai_arrayOffsets;

    /**
     * The array contains the variable area
     */
    protected int[] ai_variables;

    /**
     * The array contains instructions of a program
     */
    protected byte[] ab_programm;

    /**
     * The PC register
     */
    protected int i_regPC;

    /**
     * The SP register
     */
    protected int i_regSP;

    /**
     * The RS register
     */
    protected int i_regRS;

    /**
     * The array contains addresses of user's quick commands
     */
    protected int[] ai_quickCommands;

    /**
     * The index of the first user's command
     */
    protected int i_firstUserCommandIndex;

    /**
     * System flags of the compiled script file
     */
    protected int i_Flags;

    /**
     * The array contains string data
     */
    protected String[] as_StringArray;

    //------------------------------------

    /**
     * The pointer to the RRGLVMListener object
     */
    private RRGGLVMListener p_listener;

    //System flags
    public static final int FLAG_DEBUGINFO = 0x01;
    public static final int FLAG_COMMANDCODEOPTIMIZED = 0x02;

    //#if DEBUG
    // The sources name table
    private String[] as_sourcesTable;
    //#endif

    // The flag shows that the program contains debug info
    protected boolean lg_containsDebugInfo;

    /**
     * Constructor
     *
     * @param _inStream the input stream with byte code
     * @param _listener the listener of events
     * @throws IOException will be throwed if any error
     */
    public RRGGLbinVM(InputStream _inStream, RRGGLVMListener _listener) throws IOException
    {
        DataInputStream p_dis = new DataInputStream(_inStream);

        //---------------Read header----------------------
        // ID
        if (p_dis.readInt() != 0xCABAB0CA) throw new IOException("It is not a RRGGL file");
        //Version
        if (p_dis.readShort() != 0x0100) throw new IOException("Bad version");

        // System flags
        int i_flags = p_dis.readUnsignedByte();
        i_Flags = i_flags;

        //#if DEBUG
        lg_containsDebugInfo = (i_flags & FLAG_DEBUGINFO) != 0;
        //#endif

        //The variable area size (2 bytes) in cells
        int i_varNumber = p_dis.readUnsignedShort();
        ai_variables = new int[i_varNumber];

        //Array's number
        int i_arrNumber = p_dis.readByte();
        ai_arrayOffsets = new int[i_arrNumber];
        ash_arrayDimensions = new short[i_arrNumber << 2];

        int i_offset = 0;

        for (int li = 0; li < i_arrNumber; li++)
        {
            // The dimension number
            int i_dim = p_dis.readUnsignedByte();
            int i_indx = li << 2;
            int i_size = 1;
            ash_arrayDimensions[i_indx++] = (byte) i_dim;

            ai_arrayOffsets[li] = i_offset;

            final short [] ash_arrDim = ash_arrayDimensions;

            for (int lt = 0; lt < i_dim; lt++)
            {
                int i_val = p_dis.readUnsignedShort();
                ash_arrDim[i_indx++] = (short) i_val;
                i_size *= i_val;
            }
            i_offset += i_size;
        }

        ai_arrays = new int[i_offset];

        // Loading of string constants
        int i_number = p_dis.readUnsignedShort();
        as_StringArray = new String[i_number];
        final String [] as_strings = as_StringArray;
        for (int li = 0; li < i_number; li++)
        {
            as_strings[li] = p_dis.readUTF();
        }

        // The user words table, it is the words that can be used via short call
        // The quick words number
        int i_quickWordsNumer = p_dis.readUnsignedByte();
        ai_quickCommands = new int[i_quickWordsNumer];
        // The first word code
        i_firstUserCommandIndex = p_dis.readUnsignedByte();
        // The table
        for (int li = 0; li < i_quickWordsNumer; li++)
        {
            ai_quickCommands[li] = p_dis.readInt();
        }

        // The programm (instruction set)
        // The size
        int i_size = p_dis.readInt();
        ab_programm = new byte[i_size];

        // The data
        int i_doff = 0;
        int i_dlen = i_size;

        while (i_dlen > 0)
        {
            int i_read = p_dis.read(ab_programm, i_doff, i_dlen);
            i_dlen -= i_read;
            i_doff += i_read;
        }

        //#if DEBUG
        // Loading the sources name table
        if (lg_containsDebugInfo)
        {
            i_size = p_dis.readUnsignedShort();
            as_sourcesTable = new String[i_size];
            for (int li = 0; li < as_sourcesTable.length; li++)
            {
                as_sourcesTable[li] = p_dis.readUTF();
            }
        }
        //#endif

        //---------------Initialization----------------------------
        ai_rs = new int[SIZE_RS];
        ai_sp = new int[SIZE_SP];
        resetVM();

        p_listener = _listener;
    }

    /**
     * To get the flags of played file
     *
     * @return the flags value of the loaded script
     */
    public final int getFlags()
    {
        return i_Flags;
    }

    /**
     * Get array contains SP stack
     *
     * @return the pinter to the SP stack array
     */
    public final int[] getSPStack()
    {
        return ai_sp;
    }

    /**
     * Get array contains RS stack
     *
     * @return  the pointer to the RS stack
     */
    public final int[] getRSStack()
    {
        return ai_rs;
    }

    /**
     * Reset the VM
     */
    public final void resetVM()
    {
        i_regPC = 0;
        i_regRS = 0;
        i_regSP = 0;

        for (int li = 0; li < ai_arrays.length; li++) ai_arrays[li] = 0;
        for (int li = 0; li < ai_variables.length; li++) ai_variables[li] = 0;
    }

    /**
     * To return the state of a system register for index
     *
     * @param _register the index of a system register
     * @return the value from a system register
     */
    public final int getRegisterState(int _register)
    {
        switch (_register)
        {
            case REGISTER_PC:
                return i_regPC;
            case REGISTER_RS:
                return i_regRS;
            case REGISTER_SP:
                return i_regSP;
            default:
                throw new RuntimeException("Wrong index " + _register);
        }
    }

    /**
     * Set the value to a system register
     *
     * @param _register the index of a system register
     * @param _state    the new value for the register
     */
    public final void setRegisterState(int _register, int _state)
    {
        switch (_register)
        {
            case REGISTER_PC:
                i_regPC = _state;
                break;
            case REGISTER_RS:
                i_regRS = _state;
                break;
            case REGISTER_SP:
                i_regSP = _state;
                break;
            default:
                throw new RuntimeException("Wrong index " + _register);
        }
    }

    /**
     * Push a value to the SP
     *
     * @param _value the value
     */
    public void pushSP(int _value)
    {
        ai_sp[i_regSP++] = _value;
    }

    /**
     * Pop a value from the SP
     *
     * @return the value from the top of the SP
     */
    public final int popSP()
    {
        return ai_sp[--i_regSP];
    }

    /**
     * Push a value to the top of the RS
     *
     * @param _value the value
     */
    public final void pushRS(int _value)
    {
        ai_rs[i_regRS++] = _value;
    }

    /**
     * Pop a value from the top of RS
     *
     * @return the value from the top of the RS
     */
    public final int popRS()
    {
        return ai_rs[--i_regRS];
    }

    /**
     * To get a string from the string pool for an index
     *
     * @param _index the index of a string
     * @return a string for the index or null if the index is not in the array bounds
     */
    public final String getStringForIndex(int _index)
    {
        if (_index < 0 || _index >= as_StringArray.length) return null;
        return as_StringArray[_index];
    }

    /**
     * To set a string to the string pool for an index
     *
     * @param _index  the index of a pool cell
     * @param _string new value of the cell in the pool
     */
    public final void setStringForIndex(int _index, String _string)
    {
        if (_index < 0 || _index >= as_StringArray.length) return;
        as_StringArray[_index] = _string;
    }

    /**
     * To execute a woord for the offset
     *
     * @param _wordOffset the offset of the word in programm array
     * @param _verbose    to print information about executed position during playing of debugged script
     */
    public final void executeWord(int _wordOffset, boolean _verbose)
    {
        int i_execCounter = 0;
        i_regPC = _wordOffset;

        boolean lg_process = true;

        int i_sourceFileIndex = 0;
        int i_stringIndex = 0;
        int i_stringPos = 0;

        try
        {

            final int[] ai_sp = this.ai_sp;
            final int[] ai_rs = this.ai_rs;
            final byte[] ab_programm = this.ab_programm;
            final short [] ash_arrayDimensions = this.ash_arrayDimensions;
            final int [] ai_arrays = this.ai_arrays;
            final int [] ai_quickCommands = this.ai_quickCommands;

            while (lg_process)
            {
                //#if DEBUG
                if (lg_containsDebugInfo)
                {
                    i_sourceFileIndex = ab_programm[i_regPC++] & 0xFF;
                    int i_b0 = ab_programm[i_regPC++] & 0xFF;
                    int i_b1 = ab_programm[i_regPC++] & 0xFF;
                    i_stringIndex = (i_b0 << 8) | i_b1;

                    i_b0 = ab_programm[i_regPC++] & 0xFF;
                    i_b1 = ab_programm[i_regPC++] & 0xFF;
                    i_stringPos = (i_b0 << 8) | i_b1;

                    if (_verbose)
                    {
                       System.out.println("Exec. " + as_sourcesTable[i_sourceFileIndex] + "[" + i_stringIndex + ':' + i_stringPos + "]");
                    }
                }
                //#endif

                int i_cmnd = ab_programm[i_regPC++];

                //#if DEBUG
                if (lg_containsDebugInfo && _verbose)
                {
                    System.out.println("CODE "+(i_regPC-1)+':'+i_cmnd);
                }
                //#endif

                // Instruction processing
                switch (i_cmnd)
                {
                    //#if P_RRGGL_CMND_ADD
                    case RRGGL_CMND_ADD:
                    {
                        // SP: a b --> c where c = a+b
                        final int i_b = ai_sp[--i_regSP];
                        final int i_a = ai_sp[--i_regSP];
                        ai_sp[i_regSP++] = i_a + i_b;
                    }
                    break;
                    //#endif

                    //#if P_RRGGL_CMND_AND
                    case RRGGL_CMND_AND:
                    {
                        // SP: a b --> c where c = a and b
                        final int i_b = ai_sp[--i_regSP];
                        final int i_a = ai_sp[--i_regSP];
                        ai_sp[i_regSP++] = i_a & i_b;
                    }
                    break;
                    //#endif

                    //#if P_RRGGL_CMND_ABS
                    case RRGGL_CMND_ABS:
                    {
                        // SP: a --> b where b = ABS(a)
                        final int i_a = ai_sp[--i_regSP];
                        ai_sp[i_regSP++] = Math.abs(i_a);
                    }
                    break;
                    //#endif

                    //#if P_RRGGL_CMND_ASKBRANCH
                    case RRGGL_CMND_ASKBRANCH:
                    {
                        // SP: a -->
                        // Jump if zero
                        int i_a = ai_sp[--i_regSP];

                        if (i_a == 0)
                        {
                            i_a = ab_programm[i_regPC++];
                            final int i_b = ab_programm[i_regPC++] & 0xFF;
                            final short sh_a = (short) ((i_a << 8) | i_b);
                            i_regPC = (i_regPC - 3) + sh_a;

                            //#if DEBUG
                            if (lg_containsDebugInfo)
                            {
                                i_regPC -= DEBUG_INFO_LENGTH;
                            }
                            //#endif
                        }
                        else
                        {
                            i_regPC += 2;
                        }
                    }
                    break;
                    //#endif

                    //#if P_RRGGL_CMND_ASKDUP
                    case RRGGL_CMND_ASKDUP:
                    {
                        // SP: a --> a a  if a != 0
                        final int i_a = ai_sp[--i_regSP];
                        if (i_a == 0)
                        {
                            ai_sp[i_regSP++] = i_a;
                        }
                        else
                        {
                            ai_sp[i_regSP++] = i_a;
                            ai_sp[i_regSP++] = i_a;
                        }
                    }
                    break;
                    //#endif

                    //#if P_RRGGL_CMND_BRANCH
                    case RRGGL_CMND_BRANCH:
                    {
                        // SP: -->
                        final int i_a = ab_programm[i_regPC++];
                        final int i_b = ab_programm[i_regPC++] & 0xFF;
                        final short sh_a = (short) ((i_a << 8) | i_b);
                        i_regPC = (i_regPC - 3) + sh_a;

                        //#if DEBUG
                        if (lg_containsDebugInfo)
                        {
                            i_regPC -= DEBUG_INFO_LENGTH;
                        }
                        //#endif
                    }
                    break;
                    //#endif

                    //#if P_RRGGL_CMND_CALLOBJ
                    case RRGGL_CMND_CALLOBJ:
                    {
                        p_listener.callObject(this);
                    }
                    break;
                    //#endif

                    //#if P_RRGGL_CMND_COPY_R
                    case RRGGL_CMND_COPY_R:
                    {
                        // SP: -> a
                        // RS: a -> a
                        final int i_a = ai_rs[i_regRS - 1];
                        ai_sp[i_regSP++] = i_a;
                    }
                    break;
                    //#endif

                    //#if P_RRGGL_CMND_DEC
                    case RRGGL_CMND_DEC:
                    {
                        // SP: a --> b where b = a -1
                        final int i_a = ai_sp[--i_regSP];
                        ai_sp[i_regSP++] = i_a - 1;
                    }
                    break;
                    //#endif

                    //#if P_RRGGL_CMND_DIV
                    case RRGGL_CMND_DIV:
                    {
                        // SP: a b -> c where c = a / b
                        final int i_b = ai_sp[--i_regSP];
                        final int i_a = ai_sp[--i_regSP];
                        ai_sp[i_regSP++] = i_a / i_b;
                    }
                    break;
                    //#endif

                    //#if P_RRGGL_CMND_DROP
                    case RRGGL_CMND_DROP:
                    {
                        // SP: a ->
                        //#if DEBUG
                        if (i_regSP == 0) throw new RuntimeException("Stack empty");
                        //#endif
                        i_regSP--;
                    }
                    break;
                    //#endif

                    //#if P_RRGGL_CMND_DROPALL
                    case RRGGL_CMND_DROPALL:
                    {
                        // SP: a b c d e ... ->
                        i_regSP = 0;
                    }
                    break;
                    //#endif

                    //#if P_RRGGL_CMND_DUP
                    case RRGGL_CMND_DUP:
                    {
                        // SP: a -> a a
                        final int i_a = ai_sp[--i_regSP];
                        ai_sp[i_regSP++] = i_a;
                        ai_sp[i_regSP++] = i_a;
                    }
                    break;
                    //#endif

                    //#if P_RRGGL_CMND_EQU
                    case RRGGL_CMND_EQU:
                    {
                        // SP: a b -> c c = a == b
                        final int i_b = ai_sp[--i_regSP];
                        final int i_a = ai_sp[--i_regSP];
                        ai_sp[i_regSP++] = (i_a == i_b) ? -1 : 0;
                    }
                    break;
                    //#endif

                    //#if P_RRGGL_CMND_EQUZERO
                    case RRGGL_CMND_EQUZERO:
                    {
                        // SP: a -> b b = a == 0
                        final int i_a = ai_sp[--i_regSP];
                        ai_sp[i_regSP++] = (i_a == 0) ? -1 : 0;
                    }
                    break;
                    //#endif

                    //#if P_RRGGL_CMND_EXEC
                    case RRGGL_CMND_EXEC:
                    {
                        // SP: -->
                        // RS: --> PC of next instruction
                        // PC: a
                        final int i_b0 = ab_programm[i_regPC++] & 0xFF;
                        final int i_b1 = ab_programm[i_regPC++] & 0xFF;
                        final int i_b2 = ab_programm[i_regPC++] & 0xFF;
                        final int i_b3 = ab_programm[i_regPC++] & 0xFF;

                        int i_addr = (i_b0 << 24) | (i_b1 << 16) | (i_b2 << 8) | i_b3;

                        // place current PC on RS
                        ai_rs[i_regRS++] = i_regPC;
                        i_regPC = i_addr;
                        i_execCounter++;
                    }
                    break;
                    //#endif

                    //#if P_RRGGL_CMND_EXECUTE
                    case RRGGL_CMND_EXECUTE:
                    {
                        // SP: a -->
                        // RS: --> PC of next instruction
                        // PC: a
                        final int i_a = ai_sp[--i_regSP];
                        ai_rs[i_regRS++] = i_regPC;
                        i_regPC = i_a;
                        i_execCounter++;
                    }
                    break;
                    //#endif

                    //#if P_RRGGL_CMND_FROM_R
                    case RRGGL_CMND_FROM_R:
                    {
                        // SP: --> a
                        //RS: a -->
                        final int i_a = ai_rs[--i_regRS];
                        ai_sp[i_regSP++] = i_a;
                    }
                    break;
                    //#endif

                    //#if P_RRGGL_CMND_INC
                    case RRGGL_CMND_INC:
                    {
                        // SP: a --> b b = a + 1
                        final int i_a = ai_sp[--i_regSP];
                        ai_sp[i_regSP++] = i_a + 1;
                    }
                    break;
                    //#endif

                    //#if P_RRGGL_CMND_LESS
                    case RRGGL_CMND_LESS:
                    {
                        // SP: a b -> c c = a < b
                        final int i_b = ai_sp[--i_regSP];
                        final int i_a = ai_sp[--i_regSP];
                        ai_sp[i_regSP++] = (i_a < i_b) ? -1 : 0;
                    }
                    break;
                    //#endif

                    //#if P_RRGGL_CMND_LITI
                    case RRGGL_CMND_LITI:
                    {
                        // SP: --> a
                        final int i_b0 = ab_programm[i_regPC++] & 0xFF;
                        final int i_b1 = ab_programm[i_regPC++] & 0xFF;
                        final int i_b2 = ab_programm[i_regPC++] & 0xFF;
                        final int i_b3 = ab_programm[i_regPC++] & 0xFF;

                        final int i_val = (i_b0 << 24) | (i_b1 << 16) | (i_b2 << 8) | i_b3;

                        ai_sp[i_regSP++] = i_val;
                    }
                    break;
                    //#endif

                    //#if P_RRGGL_CMND_LITS
                    case RRGGL_CMND_LITS:
                    {
                        // SP: --> a
                        final int i_b0 = ab_programm[i_regPC++] & 0xFF;
                        final int i_b1 = ab_programm[i_regPC++] & 0xFF;

                        final short sh_val = (short) ((i_b0 << 8) | i_b1);

                        ai_sp[i_regSP++] = sh_val;
                    }
                    break;
                    //#endif

                    //#if P_RRGGL_CMND_LITB
                    case RRGGL_CMND_LITB:
                    {
                        // SP: --> a
                        final byte b_b0 = ab_programm[i_regPC++];

                        ai_sp[i_regSP++] = b_b0;
                    }
                    break;
                    //#endif

                    //#if P_RRGGL_CMND_MOD
                    case RRGGL_CMND_MOD:
                    {
                        // SP: a b -> c c = a % b
                        final int i_b = ai_sp[--i_regSP];
                        final int i_a = ai_sp[--i_regSP];
                        ai_sp[i_regSP++] = i_a % i_b;
                    }
                    break;
                    //#endif

                    //#if P_RRGGL_CMND_MORE
                    case RRGGL_CMND_MORE:
                    {
                        // SP: a b -> c c = a > b
                        final int i_b = ai_sp[--i_regSP];
                        final int i_a = ai_sp[--i_regSP];
                        ai_sp[i_regSP++] = (i_a > i_b) ? -1 : 0;
                    }
                    break;
                    //#endif

                    //#if P_RRGGL_CMND_MUL
                    case RRGGL_CMND_MUL:
                    {
                        // SP: a b -> c c = a * b
                        final int i_b = ai_sp[--i_regSP];
                        final int i_a = ai_sp[--i_regSP];
                        ai_sp[i_regSP++] = i_a * i_b;
                    }
                    break;
                    //#endif

                    //#if P_RRGGL_CMND_MAX
                    case RRGGL_CMND_MAX:
                    {
                        // SP: a b -> c c = a | b
                        final int i_b = ai_sp[--i_regSP];
                        final int i_a = ai_sp[--i_regSP];
                        ai_sp[i_regSP++] = Math.max(i_a, i_b);
                    }
                    break;
                    //#endif

                    //#if P_RRGGL_CMND_MIN
                    case RRGGL_CMND_MIN:
                    {
                        // SP: a b -> c c = a | b
                        final int i_b = ai_sp[--i_regSP];
                        final int i_a = ai_sp[--i_regSP];
                        ai_sp[i_regSP++] = Math.min(i_a, i_b);
                    }
                    break;
                    //#endif

                    //#if P_RRGGL_CMND_NOT
                    case RRGGL_CMND_NOT:
                    {
                        // SP: a -> b b = not a
                        final int i_a = ai_sp[--i_regSP];
                        ai_sp[i_regSP++] = ~i_a;
                    }
                    break;
                    //#endif

                    //#if P_RRGGL_CMND_NOOP
                    case RRGGL_CMND_NOOP:
                        break;
                        //#endif

                        //#if P_RRGGL_CMND_NOTEQU
                    case RRGGL_CMND_NOTEQU:
                    {
                        // SP: a b -> c c = a <> b
                        final int i_b = ai_sp[--i_regSP];
                        final int i_a = ai_sp[--i_regSP];
                        ai_sp[i_regSP++] = (i_a != i_b) ? -1 : 0;
                    }
                    break;
                    //#endif

                    //#if P_RRGGL_CMND_OR
                    case RRGGL_CMND_OR:
                    {
                        // SP: a b -> c c = a or b
                        final int i_b = ai_sp[--i_regSP];
                        final int i_a = ai_sp[--i_regSP];
                        ai_sp[i_regSP++] = i_a | i_b;
                    }
                    break;
                    //#endif

                    //#if P_RRGGL_CMND_OVER
                    case RRGGL_CMND_OVER:
                    {
                        // SP: a b ->  a b a
                        final int i_a = ai_sp[i_regSP - 2];
                        ai_sp[i_regSP++] = i_a;
                    }
                    break;
                    //#endif

                    //#if P_RRGGL_CMND_READ_VAR
                    case RRGGL_CMND_READ_VAR:
                    {
                        // SP: a -> b  a - index, b - value
                        final int i_a = ai_sp[--i_regSP];
                        ai_sp[i_regSP++] = ai_variables[i_a];
                    }
                    break;
                    //#endif

                    //#if P_RRGGL_CMND_ROT
                    case RRGGL_CMND_ROT:
                    {
                        // SP: a b c ->  b c a
                        final int i_a = ai_sp[i_regSP - 3];
                        final int i_b = ai_sp[i_regSP - 2];
                        final int i_c = ai_sp[i_regSP - 1];

                        ai_sp[i_regSP - 3] = i_b;
                        ai_sp[i_regSP - 2] = i_c;
                        ai_sp[i_regSP - 1] = i_a;
                    }
                    break;
                    //#endif

                    //#if P_RRGGL_CMND_RET
                    case RRGGL_CMND_RET:
                    {
                        // SP: -->
                        // RS: a -->
                        // PC: --> a
                        // If it is the last RET then we can not change RS but return from the processing cycle
                        i_execCounter--;
                        if (i_execCounter < 0)
                        {
                            lg_process = false;
                            continue;
                        }
                        final int i_pc = ai_rs[--i_regRS];
                        i_regPC = i_pc;
                    }
                    break;
                    //#endif

                    //#if P_RRGGL_CMND_READ_FROM_REG
                    case RRGGL_CMND_READ_FROM_REG:
                    {
                        //SP: a --> b     a - index b - state
                        final int i_a = ai_sp[--i_regSP];
                        final int i_state = getRegisterState(i_a);
                        ai_sp[i_regSP++] = i_state;
                    }
                    break;
                    //#endif

                    //#if P_RRGGL_CMND_READ_FROM_ARAY
                    case RRGGL_CMND_READ_FROM_ARRAY:
                    {
                        // indxN indx3 indx2 indx1 arrayIndx -> value
                        int i_indxArray = ai_sp[--i_regSP];

                        int i_dataIndex = i_indxArray << 2;
                        int i_dimNumber = ash_arrayDimensions[i_dataIndex++];
                        int i_arrayOffset = ai_arrayOffsets[i_indxArray];

                        int i_inOffset = 0;

                        int i_sz = 1;

                        for (int li = 0; li < i_dimNumber; li++)
                        {
                            int i_indx = ai_sp[--i_regSP];
                            int i_cdim = ash_arrayDimensions[li] & 0xFF;
                            i_inOffset += (i_sz * i_indx);
                            i_sz *= i_cdim;
                        }

                        int i_value = ai_arrays[i_inOffset + i_arrayOffset];

                        ai_sp[i_regSP++] = i_value;

                        // TODO it's would be good idea to insert code for test of the array bounds but it will decrease the processing speed
                    }
                    break;
                    //#endif

                    //#if P_RRGGL_CMND_SHL
                    case RRGGL_CMND_SHL:
                    {
                        // SP: a ->  b b = a * 2
                        int i_a = ai_sp[--i_regSP];
                        ai_sp[i_regSP++] = i_a << 1;
                    }
                    break;
                    //#endif

                    //#if P_RRGGL_CMND_SHR
                    case RRGGL_CMND_SHR:
                    {
                        // SP: a ->  b b = a / 2
                        int i_a = ai_sp[--i_regSP];
                        ai_sp[i_regSP++] = i_a >> 1;
                    }
                    break;
                    //#endif

                    //#if P_RRGGL_CMND_SUB
                    case RRGGL_CMND_SUB:
                    {
                        // SP: a b ->  c  c = a - b
                        int i_b = ai_sp[--i_regSP];
                        int i_a = ai_sp[--i_regSP];
                        ai_sp[i_regSP++] = i_a - i_b;
                    }
                    break;
                    //#endif

                    //#if P_RRGGL_CMND_SWAP
                    case RRGGL_CMND_SWAP:
                    {
                        // SP: a b -> b a
                        int i_b = ai_sp[--i_regSP];
                        int i_a = ai_sp[--i_regSP];
                        ai_sp[i_regSP++] = i_b;
                        ai_sp[i_regSP++] = i_a;
                    }
                    break;
                    //#endif

                    //#if P_RRGGL_CMND_SWITCH
                    case RRGGL_CMND_SWITCH:
                    {
                        // SP: index ->     index [0..255]
                        int i_indx = ai_sp[--i_regSP];
                        int i_pcSw = i_regPC - 1;
                        int i_maxIndx = ab_programm[i_regPC++] & 0xFF;
                        int i_b0 = ab_programm[i_regPC++] & 0xFF;
                        int i_b1 = ab_programm[i_regPC++] & 0xFF;
                        int i_default = (i_b0 << 8) | i_b1;
                        if (i_indx < 0 || i_indx > i_maxIndx)
                        {
                            i_regPC = i_pcSw + i_default;
                        }
                        else
                        {
                            i_regPC += (i_indx << 1);
                            i_b0 = ab_programm[i_regPC++] & 0xFF;
                            i_b1 = ab_programm[i_regPC++] & 0xFF;
                            int i_offset = (i_b0 << 8) | i_b1;
                            i_regPC = i_pcSw + i_offset;
                        }

                        //#if DEBUG
                        if (lg_containsDebugInfo)
                        {
                            i_regPC -= DEBUG_INFO_LENGTH;
                        }
                        //#endif
                    }
                    break;
                    //#endif

                    //#if P_RRGGL_CMND_SIGNAL
                    case RRGGL_CMND_SIGNAL:
                    {
                        p_listener.processSignal(this);
                    }
                    break;
                    //#endif

                    //#if P_RRGGL_CMND_TO_R
                    case RRGGL_CMND_TO_R:
                    {
                        // SP: a -->
                        // RS: --> a
                        final int i_a = ai_sp[--i_regSP];
                        ai_rs[i_regRS++] = i_a;
                    }
                    break;
                    //#endif

                    //#if P_RRGGL_CMND_WRITE_TO_ARRAY
                    case RRGGL_CMND_WRITE_TO_ARRAY:
                    {
                        // SP: indxN indx2 indx1 value array_index ->
                        int i_indxArray = ai_sp[--i_regSP];
                        int i_value = ai_sp[--i_regSP];

                        int i_dataIndex = i_indxArray << 2;
                        int i_dimNumber = ash_arrayDimensions[i_dataIndex++];
                        int i_arrayOffset = ai_arrayOffsets[i_indxArray];

                        int i_inOffset = 0;

                        int i_sz = 1;

                        for (int li = 0; li < i_dimNumber; li++)
                        {
                            int i_indx = ai_sp[--i_regSP];
                            int i_cdim = ash_arrayDimensions[li] & 0xFF;
                            i_inOffset += (i_sz * i_indx);
                            i_sz *= i_cdim;
                        }

                        ai_arrays[i_inOffset + i_arrayOffset] = i_value;
                    }
                    break;
                    //#endif

                    //#if P_RRGGL_CMND_WRITE_VAR
                    case RRGGL_CMND_WRITE_VAR:
                    {
                        // SP: a b -->      a - value b - index
                        int i_b = ai_sp[--i_regSP];
                        int i_a = ai_sp[--i_regSP];
                        ai_variables[i_b] = i_a;
                    }
                    break;
                    //#endif

                    //#if P_RRGGL_CMND_WRITE_TO_REG
                    case RRGGL_CMND_WRITE_TO_REG:
                    {
                        //SP: a b -->      a - state  b - index
                        int i_b = ai_sp[--i_regSP];
                        int i_a = ai_sp[--i_regSP];
                        setRegisterState(i_b, i_a);
                    }
                    break;
                    //#endif

                    //#if P_RRGGL_CMND_XOR
                    case RRGGL_CMND_XOR:
                    {
                        // SP: a b -> c c  = a xor b
                        int i_b = ai_sp[--i_regSP];
                        int i_a = ai_sp[--i_regSP];
                        ai_sp[i_regSP++] = i_a ^ i_b;
                    }
                    break;
                    //#endif

                    default:
                    {
                        // Processing quick call user words
                        int i_index = i_cmnd - i_firstUserCommandIndex;
                        int i_offset = ai_quickCommands[i_index];
                        ai_rs[i_regRS++] = i_regPC;
                        i_regPC = i_offset;
                        i_execCounter++;
                    }
                }
            }
        }
        catch (Exception _ex)
        {
            //#if DEBUG
            if (lg_containsDebugInfo)
            {
                String s_source = as_sourcesTable[i_sourceFileIndex];
                if (p_listener != null) p_listener.error(this, s_source, i_stringIndex, i_stringPos, _ex.getMessage());
            }
            else
            {
                _ex.printStackTrace();
            }
            //#endif
        }
    }
}
